{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "puzzle_input_lines = open(\"puzzle_input/day24.txt\").readlines()\n",
    "\n",
    "\n",
    "test_input_lines = \"\"\"Immune System:\n",
    "17 units each with 5390 hit points (weak to radiation, bludgeoning) with an attack that does 4507 fire damage at initiative 2\n",
    "989 units each with 1274 hit points (immune to fire; weak to bludgeoning, slashing) with an attack that does 25 slashing damage at initiative 3\n",
    "\n",
    "Infection:\n",
    "801 units each with 4706 hit points (weak to radiation) with an attack that does 116 bludgeoning damage at initiative 1\n",
    "4485 units each with 2961 hit points (immune to radiation; weak to fire, cold) with an attack that does 12 slashing damage at initiative 4\"\"\".split(\"\\n\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {
    "scrolled": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "14854\n"
     ]
    }
   ],
   "source": [
    "from collections import defaultdict\n",
    "import re\n",
    "\n",
    "# I don't like using classes to solve these problems, but this was the simplest way\n",
    "class UnitGroup:\n",
    "    def __init__(self, identifier, army, unit_group, boost=0):\n",
    "        units, hit_points, strengths, weaknesses, damage, damage_type, initiative = unit_group\n",
    "        self.identifier = f\"{army}.{identifier}\"\n",
    "        self.army = army\n",
    "        self.units = units\n",
    "        self.hit_points = hit_points\n",
    "        self.strengths = strengths\n",
    "        self.weaknesses = weaknesses\n",
    "        self.damage = damage + boost\n",
    "        self.damage_type = damage_type\n",
    "        self.initiative = initiative\n",
    "        \n",
    "    def effective_power(self):\n",
    "        return self.units * self.damage\n",
    "    \n",
    "    def select_target(self, unit_groups, targets):\n",
    "        highest_damage = 0\n",
    "        selected_target = None\n",
    "\n",
    "        for other in sorted(unit_groups, key=effective_power_and_initiative, reverse=True):\n",
    "            if self.army == other.army or other.identifier in targets.values() or other.units < 1:\n",
    "                continue\n",
    "\n",
    "            damage = self.calculate_damage(other)\n",
    "            if damage > highest_damage:\n",
    "                selected_target = other.identifier\n",
    "                highest_damage = damage\n",
    "\n",
    "        return selected_target\n",
    "    \n",
    "    def calculate_damage(self, enemy):\n",
    "        total_damage = self.units * self.damage\n",
    "        \n",
    "        if enemy.strengths and self.damage_type in enemy.strengths:\n",
    "            total_damage *= 0\n",
    "\n",
    "        if enemy.weaknesses and self.damage_type in enemy.weaknesses:\n",
    "            total_damage *= 2\n",
    "\n",
    "        return total_damage\n",
    "    \n",
    "    def attack(self, enemy):\n",
    "        if self.units < 1:\n",
    "            return\n",
    "        \n",
    "        total_damage = self.calculate_damage(enemy)\n",
    "        killed_units = min(enemy.units, total_damage // enemy.hit_points)\n",
    "        enemy.units -= killed_units\n",
    "        \n",
    "        return killed_units\n",
    "\n",
    "\n",
    "def parse_strengths_weaknesses(weakness_str):\n",
    "    strengths, weaknesses = None, None\n",
    "    \n",
    "    if weakness_str is not None:\n",
    "        weakness_str = weakness_str.lstrip(\"(\")\n",
    "        weakness_str = weakness_str.rstrip(\")\")        \n",
    "        parts = weakness_str.split(\"; \")\n",
    "        weak_str = \"weak to\"\n",
    "        immune_str = \"immune to\"\n",
    "        for part in parts:\n",
    "            if part.startswith(weak_str):\n",
    "                weaknesses = part.lstrip(weak_str).split(\", \")\n",
    "            elif part.startswith(immune_str):\n",
    "                strengths = part.lstrip(immune_str).split(\", \")\n",
    "\n",
    "    return strengths, weaknesses\n",
    "\n",
    "\n",
    "def parse_input_line(input_line):\n",
    "    if army_match := re.match(\"^(.+)\\:$\", input_line):\n",
    "        return army_match.group(1), None\n",
    "\n",
    "    elif unit_group_match := re.match(\"^(\\d+) units each with (\\d+) hit points (\\(.+\\))?\\s?with an attack that does (\\d+) (\\w+) damage at initiative (\\d+)$\", input_line):\n",
    "        units = int(unit_group_match.group(1))\n",
    "        hit_points = int(unit_group_match.group(2))\n",
    "        strengths, weaknesses = parse_strengths_weaknesses(unit_group_match.group(3))\n",
    "        damage = int(unit_group_match.group(4))\n",
    "        damage_type = unit_group_match.group(5)\n",
    "        initiative = int(unit_group_match.group(6))            \n",
    "        return None, (units, hit_points, strengths, weaknesses, damage, damage_type, initiative)\n",
    "    \n",
    "    else:\n",
    "        return None, None\n",
    "\n",
    "\n",
    "def parse_input(input_lines, boost=0):\n",
    "    current_army = None\n",
    "    unit_groups = []\n",
    "    for i, line in enumerate(input_lines):\n",
    "        army, unit_group = parse_input_line(line)\n",
    "        if army:\n",
    "            current_army = army\n",
    "        elif unit_group:\n",
    "            if current_army == \"Immune System\":\n",
    "                unit_groups.append(UnitGroup(i, current_army, unit_group, boost))\n",
    "            else:\n",
    "                unit_groups.append(UnitGroup(i, current_army, unit_group))\n",
    "        else:\n",
    "            pass\n",
    "    return unit_groups\n",
    "\n",
    "\n",
    "def effective_power(unit_group):\n",
    "    return unit_group.effective_power()\n",
    "\n",
    "\n",
    "def effective_power_and_initiative(unit_group):\n",
    "    return (unit_group.effective_power(), unit_group.initiative)\n",
    "\n",
    "\n",
    "def initiative(unit_group):\n",
    "    return unit_group.initiative\n",
    "\n",
    "\n",
    "def get_unit_group(unit_groups, unit_group_id):\n",
    "    for unit_group in unit_groups:\n",
    "        if unit_group.identifier == unit_group_id:\n",
    "            return unit_group\n",
    "    return None\n",
    "\n",
    "\n",
    "def armies_remaining(unit_groups):\n",
    "    armies = set()\n",
    "    \n",
    "    for unit_group in unit_groups:\n",
    "        if unit_group.units > 0:\n",
    "            armies.add(unit_group.army)\n",
    "    \n",
    "    return len(armies)\n",
    "\n",
    "\n",
    "def fight(unit_groups):\n",
    "    killed_units = 1\n",
    "    while killed_units > 0:\n",
    "        killed_units = 0\n",
    "        \n",
    "        # target selection\n",
    "        targets = {}\n",
    "        for unit_group in sorted(unit_groups, key=effective_power_and_initiative, reverse=True):\n",
    "            target = unit_group.select_target(unit_groups, targets)\n",
    "            if target is not None:\n",
    "                targets[unit_group.identifier] = target\n",
    "\n",
    "        # attacking\n",
    "        for unit_group in sorted(unit_groups, key=initiative, reverse=True):\n",
    "            if unit_group.units < 1 or unit_group.identifier not in targets:\n",
    "                continue\n",
    "\n",
    "            other = get_unit_group(unit_groups, targets[unit_group.identifier])\n",
    "            killed_units += unit_group.attack(other)\n",
    "\n",
    "            \n",
    "def part_one(input_lines):\n",
    "    unit_groups = parse_input(input_lines)\n",
    "    fight(unit_groups)\n",
    "    return sum(ug.units for ug in unit_groups)\n",
    "\n",
    "\n",
    "assert 5216 == part_one(test_input_lines)\n",
    "print(part_one(puzzle_input_lines))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "3467\n"
     ]
    }
   ],
   "source": [
    "def remaining_units(unit_groups, army):\n",
    "    total = 0\n",
    "    \n",
    "    for unit_group in unit_groups:\n",
    "        if unit_group.army == army:\n",
    "            total += unit_group.units\n",
    "            \n",
    "    return total\n",
    "\n",
    "\n",
    "def part_two(input_lines):\n",
    "    unit_groups = parse_input(input_lines)\n",
    "    boost = 1\n",
    "    fight(unit_groups)\n",
    "    \n",
    "    immune_units = remaining_units(unit_groups, \"Immune System\")\n",
    "    infection_units = remaining_units(unit_groups, \"Infection\")\n",
    "    \n",
    "    while immune_units == 0 or infection_units > 0:\n",
    "        # I had a much-faster binary search, but this is cleaner to read :)\n",
    "        boost += 1\n",
    "        unit_groups = parse_input(input_lines, boost)\n",
    "        fight(unit_groups)\n",
    "        immune_units = remaining_units(unit_groups, \"Immune System\")\n",
    "        infection_units = remaining_units(unit_groups, \"Infection\")       \n",
    "    \n",
    "    return sum(ug.units for ug in unit_groups)\n",
    "\n",
    "\n",
    "assert 51 == part_two(test_input_lines)\n",
    "print(part_two(puzzle_input_lines))"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
